from django import forms
from django.conf import settings
from django.core.paginator import Paginator, Page, EmptyPage
from django.http import Http404
from django.utils import six
from pyelasticsearch import ElasticSearch


class ESPaginator(Paginator):
    """
    Override Django's built-in Paginator class to take in a count/total number of items;
    ElasticSearch provides the total as a part of the query results, so we can minimize hits.

    Also change `page()` method to use custom ESPage class (see below).
    """
    def __init__(self, *args, **kwargs):
        count = kwargs.pop('count', None)
        super(ESPaginator, self).__init__(*args, **kwargs)
        self._count = count

    def page(self, number):
        "Returns a Page object for the given 1-based page number."
        number = self.validate_number(number)
        return ESPage(self.object_list, number, self)


class ESPage(Page):
    def __getitem__(self, index):
        if not isinstance(index, (slice,) + six.integer_types):
            raise TypeError
        obj = self.object_list[index]
        return obj.get('_source', obj)


class ESSearchForm(forms.Form):
    def get_index(self):
        # return the ES index name (or multiple comma-separated) you want to target, or None if you don't want to target an index
        raise NotImplementedError

    def get_type(self):
        # return the ES type name (or multiple comma-separated) you want to target, or None if you don't want to target a type
        raise NotImplementedError

    def prepare_query(self, count, page=1, page_size=10, **kwargs):
        # `count` is a boolean indicating whether we want to simply get
        # the count of matchin docs or the actual results
        raise NotImplementedError

    def _get_results(self, count, **kwargs):
        self.is_valid()
        es = kwargs.pop('es', ElasticSearch(settings.ES_CONNECTION_URL))

        try:
            page_size = int(kwargs.pop('page_size', 20))
        except (ValueError, TypeError):
            page_size = 20
        try:
            page = int(kwargs.pop('page', 1))
        except (ValueError, TypeError):
            page = 1

        query = self.prepare_query(count, page, page_size, **kwargs)
        if type(query) in (list, tuple):
            msearch_data = u''
            for q in query:
                data = {'index': self.get_index(), 'type': self.get_type()}
                if 'routing' in kwargs:
                    data['routing'] = kwargs.get('routing')

                # if we want to retrieve search counts, we tweak things a little
                if count:
                    data['search_type'] = 'count'

                msearch_data += es._encode_json(data) + '\n'
                msearch_data += es._encode_json(q) + '\n'

            responses = es.send_request(
                'POST',
                ['_msearch'],
                msearch_data,
                encode_body=False
            )

            results = []
            if not count:
                return responses.get('responses', [])
            else:
                for response in responses.get('responses', []):
                    hit_stats = response.pop('hits', {})
                    total = hit_stats.get('total', 0)
                    results.append(total)

            return results

        elif not count:
            query_params = {}
            if 'routing' in kwargs:
                query_params['routing'] = kwargs.get('routing')

            response = es.search(
                query,
                index=self.get_index(),
                doc_type=self.get_type(),
                # query_params=query_params
            )
            hit_stats = response.pop('hits', {})
            hit_data = hit_stats.pop('hits', [])
            if kwargs.get('paginate', True):
                paginator = ESPaginator(
                    hit_data,
                    page_size,
                    count=hit_stats.get('total', 0)
                )

                try:
                    return paginator.page(page)
                except EmptyPage:
                    raise Http404
            else:
                return hit_data
        else:
            query_params = {}
            if 'routing' in kwargs:
                query_params['routing'] = kwargs.get('routing')

            return es.count(
                query.get('query', {}),
                index=self.get_index(),
                doc_type=self.get_type(),
                # query_params=query_params
            ).get('count', 0)

    def search(self, **kwargs):
        return self._get_results(count=False, **kwargs)

    def count(self, **kwargs):
        return self._get_results(count=True, **kwargs)



from django.conf import settings


# Using this causes the ESDirectMixin and ESRabbitMQMixin to temporarily store ES bulk-formatted
# index/delete requests from save/delete signal calls until the request has been sent to the client
# and then sends them to the broker/task/ES. This ensures that client response times are affected
# as little as possible.
ES_USE_REQUEST_FINISHED_SIGNAL = getattr(settings, 'ES_USE_REQUEST_FINISHED_SIGNAL', False)
ES_BULK_LIMIT_BEFORE_SEND = getattr(settings, 'ES_BULK_LIMIT_BEFORE_SEND', 500)
ES_CONNECTION_URL = getattr(settings, 'ES_CONNECTION_URL', 'http://localhost:9200/')

# Override this if you want to have a base set of settings for all your indexes. This dictionary
# gets cloned and then updated with custom index-specific from your ES_CUSTOM_INDEX_SETTINGS
# Eg. to ensure that all of your indexes have 1 shard and have an edgengram tokenizer/analyzer
# configured
# ES_DEFAULT_INDEX_SETTINGS = {
#     "settings" : {
#         "index" : {
#             "number_of_replicas" : 1
#         },
#         "analysis" : {
#             "analyzer" : {
#                 "left" : {
#                     "filter" : [
#                         "standard",
#                         "lowercase",
#                         "stop"
#                     ],
#                     "type" : "custom",
#                     "tokenizer" : "left_tokenizer"
#                 }
#             },
#             "tokenizer" : {
#                 "left_tokenizer" : {
#                     "side" : "front",
#                     "max_gram" : 12,
#                     "type" : "edgeNGram"
#                 }
#             }
#         }
#     }
# }
ES_DEFAULT_INDEX_SETTINGS = getattr(settings, 'ES_DEFAULT_INDEX_SETTINGS', {})

# Override this in your project settings to define any ElasticSearch-specific index settings.
# Eg.
# ES_CUSTOM_INDEX_SETTINGS = {
#     "twitter": {
#         "settings" : {
#             "index" : {
#                 "number_of_shards" : 3,
#             }
#         }
#     },
#     "<your-other-index-name>": {
#         "settings" : {
#             "index" : {
#                 "number_of_shards" : 50,
#                 "number_of_replicas" : 2
#             }
#         }
#     }
# }
ES_CUSTOM_INDEX_SETTINGS = getattr(settings, 'ES_CUSTOM_INDEX_SETTINGS', {})



import collections
import gc
import inspect

from django.conf import settings
from django.http import Http404
from django.utils.importlib import import_module
from pyelasticsearch import ElasticSearch, ElasticHttpNotFoundError

try:
    import celery
except ImportError:
    celery = None

from .indexes import ESBaseIndex


all_indexes = None


def get_all_indexes(es=None):
    global all_indexes
    if not all_indexes:
        all_indexes = collections.defaultdict(lambda: [])
        for app in settings.INSTALLED_APPS:
            try:
                index_module = import_module('.es', app)
            except ImportError:
                continue

            for name, item in inspect.getmembers(index_module):
                if inspect.isclass(item) and issubclass(item, ESBaseIndex) and item is not ESBaseIndex:
                    obj = item(es=es)
                    all_indexes[obj.get_index_name()].append(obj)

    return all_indexes


def recursive_dict_update(d, u):
    for k, v in u.iteritems():
        if isinstance(v, collections.Mapping):
            r = recursive_dict_update(d.get(k, {}), v)
            d[k] = r
        else:
            d[k] = u[k]
    return d


def queryset_iterator(queryset, chunksize=1000):
    last_pk = queryset.order_by('-pk')[0].pk
    queryset = queryset.order_by('pk')
    pk = queryset[0].pk - 1
    while pk < last_pk:
        for row in queryset.filter(pk__gt=pk)[:chunksize]:
            pk = row.pk
            yield row
        gc.collect()


if celery:
    class ESCallbackTask(celery.Task):
        def _update_es(self):
            try:
                from .indexes import process_bulk_data
            except ImportError:
                # ES_USE_REQUEST_FINISHED_SIGNAL is not enabled, so nothing to do!
                return

            # trigger the sending of bulk data to the elasticsearch `bulk` API endpoint
            process_bulk_data(None)

        def on_success(self, retval, task_id, args, kwargs):
            self._update_es()

        def on_failure(self, exc, task_id, args, kwargs, einfo):
            self._update_es()


def get_from_es_or_None(index, type, id, **kwargs):
    es = kwargs.pop('es', ElasticSearch(settings.ES_CONNECTION_URL))
    try:
        return es.get(index, type, id, **kwargs)
    except ElasticHttpNotFoundError:
        return None


def get_from_es_or_404(index, type, id, **kwargs):
    item = get_from_es_or_None(index, type, id, **kwargs)
    if not item:
        raise Http404('No {0} matches the parameters.'.format(type))
    return item



import collections

from django.db.models import signals
from pyelasticsearch import ElasticSearch

from .settings import ES_USE_REQUEST_FINISHED_SIGNAL, ES_BULK_LIMIT_BEFORE_SEND, ES_CONNECTION_URL


if ES_USE_REQUEST_FINISHED_SIGNAL:
    from django.core.signals import request_finished
    ES_REQUEST_FINISHED_DATA = collections.defaultdict(lambda: [])

    def process_bulk_data(sender, **kwargs):
        global ES_REQUEST_FINISHED_DATA

        # ask each ES index instance to handle its own 'sending' to ES
        for index, data in ES_REQUEST_FINISHED_DATA.iteritems():
            if data:
                index.bulk_send(data)

        # Reset the global data object
        ES_REQUEST_FINISHED_DATA = collections.defaultdict(lambda: [])

    request_finished.connect(process_bulk_data, dispatch_uid=u'ES_USE_REQUEST_FINISHED_SIGNAL')


class ESBaseIndex(object):
    use_signals = False
    _es = None

    def __init__(self, *args, **kwargs):
        self._model = self.get_model()
        self._index_name = self.get_index_name()
        self._type_name = self.get_type_name()

        # Did we get an ElasticSearch object passed in? If so, use it
        self._es = kwargs.get('es', None)

    @property
    def es(self):
        if not self._es:
            self._es = ElasticSearch(ES_CONNECTION_URL)
        return self._es

    def register_signals(self):
        # By default, use_signals is False to disable save/delete handling - leaving the user to manually
        # update the ES index via the management command included with this app or use one of
        # the provided mixins to enable signal handling.
        if self.use_signals:
            signals.post_save.connect(
                receiver=self.handle_save,
                sender=self._model,
                dispatch_uid=u"{0}.{1}.save".format(
                    self._model._meta.app_label,
                    self._model._meta.module_name
                )
            )
            signals.post_delete.connect(
                receiver=self.handle_delete,
                sender=self._model,
                dispatch_uid=u"{0}.{1}.delete".format(
                    self._model._meta.app_label,
                    self._model._meta.module_name
                )
            )

    def handle_save(self, sender, instance, **kwargs):
        self.perform_action(instance, 'index')

    def handle_delete(self, sender, instance, **kwargs):
        self.perform_action(instance, 'delete')

    def perform_action(self, obj, operation, bulk=ES_USE_REQUEST_FINISHED_SIGNAL):
        if obj and self.get_object_id(obj):
            if operation == 'index' and self.should_index(obj):
                if bulk:
                    self.bulk_queue(obj, operation)
                else:
                    self.index_object(obj)
            elif operation == 'delete':
                if bulk:
                    self.bulk_queue(obj, operation)
                else:
                    self.delete_object(obj)

    def bulk_queue(self, obj, operation):
        global ES_REQUEST_FINISHED_DATA
        ES_REQUEST_FINISHED_DATA[self].append(self.bulk_prepare(obj, operation))

        # ensure we don't build up too big of a queue of data - we don't want
        # to eat up too much memory, so just send if we hit a threshold
        if len(ES_REQUEST_FINISHED_DATA[self]) >= ES_BULK_LIMIT_BEFORE_SEND:
            self.bulk_send(ES_REQUEST_FINISHED_DATA[self])
            ES_REQUEST_FINISHED_DATA[self] = []

    def bulk_prepare(self, obj, operation):
        """ This method is building custom ES bulk-formatted lines so that we can send a
        custom request through pyelasticsearch as its bulk_index() implementation has parameter
        limitations. This simply JSON-dumps the appropriate structures to a string.
        """
        data = {
            '_index': self._index_name,
            '_type': self._type_name,
            "_id": self.get_object_id(obj)
        }
        data.update(self.get_object_params(obj))

        # bulk operation instruction line
        op = self.es._encode_json({operation: data}) + '\n'

        # bulk operation data line
        if operation.lower() != 'delete':
            op += self.es._encode_json(self.get_object_data(obj)) + '\n'

        return op

    def bulk_send(self, data):
        # We use a custom bulk operation request due to limitations of the current
        # design of pyelasticsearch's implementation; for more details, see the
        # queue_data() method comments
        self.es.send_request(
            'POST',
            ['_bulk'],
            ''.join(data),
            encode_body=False,
            query_params=self.get_bulk_operation_params()
        )

    def index_object(self, obj):
        self.es.index(
            self._index_name,
            self._type_name,
            self.get_object_data(obj),
            id=self.get_object_id(obj),
            **self.get_object_params(obj)
        )

    def delete_object(self, obj):
        self.es.delete(
            self._index_name,
            self._type_name,
            id=self.get_object_id(obj),
            **self.get_object_params(obj)
        )

    def get_index_name(self):
        return self._model._meta.app_label.lower()

    def get_type_name(self):
        return self._model._meta.module_name.lower()

    def get_queryset(self):
        return self._model._default_manager.all()

    def should_index(self, obj):
        return True

    def get_mapping(self):
        # default to letting ES 'auto' map/schema the data fields
        return {}

    def get_bulk_operation_params(self):
        return {}

    def get_object_params(self, obj):
        # get routing, ttl, timestamp, etc. options for the item
        return {}

    def get_object_id(self, obj):
        return obj.pk

    def get_object_data(self, obj):
        # This must return a dictionary
        raise NotImplementedError

    def get_model(self):
        raise NotImplementedError


class ESDirectMixin:
    """ Use this mixin to enable auto-indexing/deleting through Django model
    save/delete signal handlers.
    """
    use_signals = True


# TODO: Consider implementing Elasticsearch 'river' mixins; example below
# class ESRabbitMQMixin:
#     """ Use this mixin to enable auto-indexing/deleting through Django model
#     save/delete signal handlers. Instead of communicating with ES directly, this
#     mixin sends all single/bulk requests through RabbitMQ using the ES 'river'
#     configuration.
#     """
#     use_signals = True
#
#     # ... other implementation overrides here ...



from setuptools import setup, find_packages

try:
    import pypandoc
    description_files = [
        'README.md',
        'HISTORY.md',
        'AUTHORS.md',
        'CONTRIBUTING.md'
    ]

    long_description = u''
    for f in description_files:
        try:
            long_description += pypandoc.convert(f, 'rst') + '\n\n'
        except IOError:
            pass
except ImportError:
    long_description = u'View `django-simple-elasticsearch documentation on Github  <https://github.com/jaddison/django-simple-elasticsearch>`_.'


setup(
    name = 'django-simple-elasticsearch',
    version='0.1.6',
    description = 'Simple Django ElasticSearch indexing integration.',
    long_description=long_description,
    url='http://github.com/jaddison/django-simple-elasticsearch',
    license='BSD',
    author='James Addison',
    author_email='code@scottisheyes.com',
    packages=find_packages(exclude=['tests*']),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: BSD License',
        'Intended Audience :: Developers',
        'Environment :: Web Environment',
        'Programming Language :: Python',
        'Framework :: Django',
        'Topic :: Internet :: WWW/HTTP :: WSGI'
    ]
)
